#include <ep9315/regmap.h>
#include <ep9315/regs_smc.h>
#include <target/io.h>

#define MULTIPLE                 2

#define TIMEOUT                  50000

#define ide_inb(port)    (*(volatile unsigned char *)(port))
#define ide_inw(port)    (*(volatile unsigned short *)(port))
#define ide_inl(port)    (*(volatile unsigned long *)(port))
#define ide_outb(b,port) (*(volatile unsigned char *)(port)=(b))
#define ide_outw(w,port) (*(volatile unsigned short *)(port)=(w))
#define ide_outl(l,port) (*(volatile unsigned long *)(port)=(l))

#define delay(count) {int d;for(d=0;d<(count*MULTIPLE);d++);}

/*
 * CIS Tupel codes
 */
#define CISTPL_NULL		0x00
#define CISTPL_DEVICE		0x01
#define CISTPL_LONGLINK_CB	0x02
#define CISTPL_INDIRECT		0x03
#define CISTPL_CONFIG_CB	0x04
#define CISTPL_CFTABLE_ENTRY_CB 0x05
#define CISTPL_LONGLINK_MFC	0x06
#define CISTPL_BAR		0x07
#define CISTPL_PWR_MGMNT	0x08
#define CISTPL_EXTDEVICE	0x09
#define CISTPL_CHECKSUM		0x10
#define CISTPL_LONGLINK_A	0x11
#define CISTPL_LONGLINK_C	0x12
#define CISTPL_LINKTARGET	0x13
#define CISTPL_NO_LINK		0x14
#define CISTPL_VERS_1		0x15
#define CISTPL_ALTSTR		0x16
#define CISTPL_DEVICE_A		0x17
#define CISTPL_JEDEC_C		0x18
#define CISTPL_JEDEC_A		0x19
#define CISTPL_CONFIG		0x1a
#define CISTPL_CFTABLE_ENTRY	0x1b
#define CISTPL_DEVICE_OC	0x1c
#define CISTPL_DEVICE_OA	0x1d
#define CISTPL_DEVICE_GEO	0x1e
#define CISTPL_DEVICE_GEO_A	0x1f
#define CISTPL_MANFID		0x20
#define CISTPL_FUNCID		0x21
#define CISTPL_FUNCE		0x22
#define CISTPL_SWIL		0x23
#define CISTPL_END		0xff

/*
 * CIS Function ID codes
 */
#define CISTPL_FUNCID_MULTI	0x00
#define CISTPL_FUNCID_MEMORY	0x01
#define CISTPL_FUNCID_SERIAL	0x02
#define CISTPL_FUNCID_PARALLEL	0x03
#define CISTPL_FUNCID_FIXED	0x04
#define CISTPL_FUNCID_VIDEO	0x05
#define CISTPL_FUNCID_NETWORK	0x06
#define CISTPL_FUNCID_AIMS	0x07
#define CISTPL_FUNCID_SCSI	0x08

/*
 * Fixed Disk FUNCE codes
 */
#define CISTPL_IDE_INTERFACE	0x01

#define CISTPL_FUNCE_IDE_IFACE	0x01
#define CISTPL_FUNCE_IDE_MASTER	0x02
#define CISTPL_FUNCE_IDE_SLAVE	0x03

/* First feature byte */
#define CISTPL_IDE_SILICON	0x04
#define CISTPL_IDE_UNIQUE	0x08
#define CISTPL_IDE_DUAL		0x10

/* Second feature byte */
#define CISTPL_IDE_HAS_SLEEP	0x01
#define CISTPL_IDE_HAS_STANDBY	0x02
#define CISTPL_IDE_HAS_IDLE	0x04
#define CISTPL_IDE_LOW_POWER	0x08
#define CISTPL_IDE_REG_INHIBIT	0x10
#define CISTPL_IDE_HAS_INDEX	0x20
#define CISTPL_IDE_IOIS16	0x40

#define	MAX_TUPEL_SZ	512
#define MAX_FEATURES	4

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned long u32;

static int pcmcia_check_ide_device (unsigned long base)
{
	volatile u8 *ident = 0;
	volatile u8 *feature_p[MAX_FEATURES];
	volatile u8 *p, *start;
	int n_features = 0;
	u8 func_id = ~0;
	u8 code, len;
	unsigned short config_base = 0;
	int found = 0;

	start = p = (volatile u8 *) base;

	while ((p - start) < MAX_TUPEL_SZ) {

		code = *p; p += 2;

		if (code == 0xFF) { /* End of chain */
			break;
		}

		len = *p; p += 2;
		switch (code) {
		case CISTPL_VERS_1:
			ident = p + 4;
			break;
		case CISTPL_FUNCID:
			func_id = *p;
			break;
		case CISTPL_FUNCE:
			if (n_features < MAX_FEATURES)
				feature_p[n_features++] = p;
			break;
		case CISTPL_CONFIG:
			config_base = (*(p+6) << 8) + (*(p+4));
		default:
			break;
		}
		p += 2 * len;
	}

	if (func_id != ((u8)~0)) {
		if (func_id == CISTPL_FUNCID_FIXED)
			found = 1;
		else {
			return (1);	/* no disk drive */
		}
	}

	if (!found) {
		return (1);
	}

	/* set level mode irq and I/O mapped device in config reg*/
	*((u8 *)(base + config_base + 6)) = 0x0;
	*((u8 *)(base + config_base + 0)) = 0x41;
	*((u8 *)(base + config_base + 2)) = 0x0;
	*((u8 *)(base + config_base + 4)) = 0x0;

	return (0);
}

#define PCMCIA_BASE_PHYS	0x40000000
#define PCMCIA_ADDR_MEM 	(PCMCIA_BASE_PHYS + 0xC000000)
#define PCMCIA_ADDR_ATTR	(PCMCIA_BASE_PHYS + 0x8000000)
#define PCMCIA_ADDR_IO		(PCMCIA_BASE_PHYS)

#define PCMCIA_WP          	0x01
#define PCMCIA_CD1         	0x02
#define PCMCIA_CD2         	0x04
#define PCMCIA_BVD1        	0x08
#define PCMCIA_BVD2        	0x10
#define PCMCIA_VS1         	0x20
#define PCMCIA_RDY         	0x40
#define PCMCIA_VS2		0x80

static int pcmcia_ide_init (void)
{
	unsigned long ulPFDR;
	unsigned long ulTemp;

	ide_outl( ide_inl(GPIO_PADR) | 0x8, GPIO_PADR );	//Power off
	delay (300000);
	ulTemp = ide_inl(GPIO_FINTEN);
	ide_outl( ulTemp & ~(PCMCIA_RDY | PCMCIA_CD1 |
			     PCMCIA_CD2 | PCMCIA_BVD2 |
			     PCMCIA_BVD1),
		  GPIO_FINTEN);
	ulTemp = ide_inl(GPIO_FINTEN);
	ulTemp = ide_inl(GPIO_PFDDR);
	ide_outl( ulTemp & ~(PCMCIA_RDY | PCMCIA_CD1 |
			     PCMCIA_CD2 | PCMCIA_BVD2 |
			     PCMCIA_BVD1),
		  GPIO_PFDDR);
	ulTemp = ide_inl(GPIO_PFDDR);
	ulTemp = ide_inl(GPIO_FDB);
	ide_outl( (ulTemp & ~(PCMCIA_RDY | PCMCIA_BVD2 |
			      PCMCIA_BVD1)) |
		  PCMCIA_CD1 | PCMCIA_CD2,
		  GPIO_FDB );
	ulTemp = ide_inl(GPIO_FINTTYPE1);
	ide_outl( (ulTemp & ~PCMCIA_RDY) |
		  PCMCIA_CD1 | PCMCIA_CD2 |
		  PCMCIA_BVD2 | PCMCIA_BVD1,
		  GPIO_FINTTYPE1 );
	ulTemp = ide_inl(GPIO_FINTTYPE2);
	ide_outl( ulTemp & ~(PCMCIA_RDY | PCMCIA_CD1 |
			     PCMCIA_CD2 | PCMCIA_BVD2 |
			     PCMCIA_BVD1),
		  GPIO_FINTTYPE2 );
	ulTemp = ide_inl(GPIO_FINTTYPE2);
	ide_outl( PCMCIA_RDY | PCMCIA_CD1 |
		  PCMCIA_CD2 | PCMCIA_BVD2 |
		  PCMCIA_BVD1,
		  GPIO_FEOI );
	ulTemp = ide_inl(GPIO_FEOI);
	ide_outl( 0x800a0405, SMC_PCIO );
	ulTemp = ide_inl(SMC_PCIO);
	ide_outl( 0x110403, SMC_PCAttribute );
	ulTemp = ide_inl(SMC_PCAttribute);
	ide_outl( 0x110403, SMC_PCCommon );
	ulTemp = ide_inl(SMC_PCCommon);
	ulTemp = (ide_inl(SMC_PCMCIACtrl) | PCCONT_WEN | PCCONT_PC1EN) &
	  ~PCCONT_PC1RST;
	ide_outl( ulTemp, SMC_PCMCIACtrl );
	ulTemp = ide_inl(SMC_PCMCIACtrl);
	ulTemp = ide_inl(GPIO_FINTEN);
	ide_outl( (ulTemp & ~(PCMCIA_CD1 | PCMCIA_CD2 |
			      PCMCIA_BVD2 | PCMCIA_BVD1)) |
		  PCMCIA_RDY,
		  GPIO_FINTEN );
	ulTemp = ide_inl(GPIO_FINTEN);
	
	ulPFDR = ide_inl(GPIO_PFDR);
	
	if (ulPFDR & (PCMCIA_CD1 | PCMCIA_CD2)) {
	  return -1;		/* No card in slot */
	}

	ide_outl( ide_inl(GPIO_PADR) & ~0x8, GPIO_PADR );	//Power on
	delay (300000);

	ulTemp = ide_inl(SMC_PCMCIACtrl) |
	  PCCONT_WEN | PCCONT_PC1EN | PCCONT_PC1RST;
	ide_outl( ulTemp, SMC_PCMCIACtrl );
	ulTemp = (ide_inl(SMC_PCMCIACtrl) | PCCONT_WEN | PCCONT_PC1EN) &
	  ~PCCONT_PC1RST;
	delay (10);
	ide_outl( ulTemp, SMC_PCMCIACtrl );
	ulTemp = ide_inl(SMC_PCMCIACtrl);
	delay (300000);
	
	if (pcmcia_check_ide_device (PCMCIA_ADDR_ATTR)) {
	  ide_outl( ide_inl(GPIO_PADR) | 0x8, GPIO_PADR );	//Power off
	  delay (300000);
	  return -1;
	}
	
	ide_outb(0x02, PCMCIA_ADDR_IO + 0xe);
	
	return 0;
}

#define IDE_DATA                 (PCMCIA_ADDR_IO + 0x00)
#define IDE_ERROR                (PCMCIA_ADDR_IO + 0x01)
#define IDE_FEAUTURES            (PCMCIA_ADDR_IO + 0x01)
#define IDE_SECTOR_COUNT         (PCMCIA_ADDR_IO + 0x02)
#define IDE_SECTOR_NUMBER        (PCMCIA_ADDR_IO + 0x03)
#define IDE_SECTOR_CYLINDER_LOW  (PCMCIA_ADDR_IO + 0x04)
#define IDE_SECTOR_CYLINDER_HIGH (PCMCIA_ADDR_IO + 0x05)
#define IDE_DEVICE_HEAD          (PCMCIA_ADDR_IO + 0x06)
#define IDE_STATUS               (PCMCIA_ADDR_IO + 0x07)
#define IDE_COMMAND              (PCMCIA_ADDR_IO + 0x07)
#define IDE_ALTERNATE_STATUS     (PCMCIA_ADDR_IO + 0x0e)
#define IDE_DEVICE_CONTROL       (PCMCIA_ADDR_IO + 0x0e)

#define IDE_STATUS_BUSY          0x80
#define IDE_STATUS_DRDY          0x40
#define IDE_STATUS_DRQ           0x08
#define IDE_STATUS_ERR           0x01

#define IDE_COMMAND_READ_SECTORS 0x20
#define IDE_COMMAND_IDLE         0xe3
#define IDE_COMMAND_IDENTIFY     0xec

#define IDE_DEVICE_CONTROL_SRST  0x04
#define IDE_DEVICE_CONTROL_NIEN  0x02

#define IDE_LBA                  0x40

static __inline__ int wait_nbusy_ndrq (void)
{
  int i;
  
  for (i = 0; i < TIMEOUT && (ide_inb (IDE_ALTERNATE_STATUS) &
			      (IDE_STATUS_BUSY | IDE_STATUS_DRQ)); i++) {
    delay (256);
  }
  if (i >= TIMEOUT) {
    return -1;
  }
  
  return 0;
}

static __inline__ int wait_nbusy (void)
{
  int i;
  
  for (i = 0; i < TIMEOUT && (ide_inb (IDE_ALTERNATE_STATUS) &
			      IDE_STATUS_BUSY); i++) {
    delay (256);
  }
  if (i >= TIMEOUT) {
    return -1;
  }
  
  return 0;
}

static __inline__ int wait_ready (void)
{
  int i;
  
  for (i = 0; i < TIMEOUT && !(ide_inb (IDE_ALTERNATE_STATUS) &
			      IDE_STATUS_DRDY); i++) {
    delay (256);
  }
  if (i >= TIMEOUT) {
    return -1;
  }
  
  return 0;
}

static int ide_reset (void)
{
  int ret;
  
  ret = pcmcia_ide_init ();
  if (ret) {
    return ret;
  }
  
  ide_outb (IDE_DEVICE_CONTROL_SRST | IDE_DEVICE_CONTROL_NIEN,
	    IDE_DEVICE_CONTROL);
  delay (40000);
  
  ide_outb (IDE_DEVICE_CONTROL_NIEN, IDE_DEVICE_CONTROL);
  delay (40000);
  
  ret = wait_nbusy_ndrq ();
  if (ret) {
    return ret;
  }
  
  if (ide_inb (IDE_ERROR) != 1) {
    return -1;
  }
  
  return 0;
}

static int ide_device_selection (int dev)
{
  int ret;
  
  ret = wait_nbusy_ndrq ();
  if (ret) {
    return ret;
  }
  
  ide_outb (IDE_DEVICE_CONTROL_NIEN, IDE_DEVICE_CONTROL);
  ide_outb (dev << 4, IDE_DEVICE_HEAD);
  delay (256);
  
  return wait_nbusy_ndrq ();
}

static int ide_register_check (void)
{
  ide_outb (0x55, IDE_SECTOR_CYLINDER_LOW);
  ide_outb (0xaa, IDE_SECTOR_CYLINDER_HIGH);
  if (ide_inb (IDE_SECTOR_CYLINDER_LOW) != 0x55) {
    return -1;
  }
  if (ide_inb (IDE_SECTOR_CYLINDER_HIGH) != 0xaa) {
    return -1;
  }
  
  return 0;
}

static int ide_get_data (unsigned char *buf, int count)
{
  int            ret, i;
  unsigned short data;
  
  while (count--) {
    ret = wait_nbusy ();
    if (ret) {
      return ret;
    }
    
    if (ide_inb (IDE_ALTERNATE_STATUS) & IDE_STATUS_ERR) {
      return -1;
    }
    
    for (i = 0; i < 256; i++) {
      data = ide_inw (IDE_DATA);
      *buf++ = data & 0xff;
      *buf++ = data >> 8;
    }
  }
  
  return 0;
}

static void ide_fix_string (unsigned char *dest, unsigned char *src, int num)
{
  int i;
  
  for (i = 0; i < num; i+= 2) {
    *dest++ = *(src + 1);
    *dest++ = *src;
    src += 2;
  }
  while (*--dest == (' '))
    ;
  ++dest;
  *dest++ = ' ';
  *dest++ = '\0';
}

static int ide_identify_device (int dev)
{
  int            ret;
  unsigned short buf[256];
  unsigned char  name[42];
  
  ret = ide_device_selection (dev);
  if (ret) {
    return ret;
  }
  
  ret = wait_ready ();
  if (ret) {
    return ret;
  }
  
  ide_outb (IDE_COMMAND_IDENTIFY, IDE_COMMAND);
  delay (256);
  
  ret = ide_get_data ((unsigned char *)buf, 1);
  if (ret) {
    return ret;
  }
  
  if (*buf & 0x04) {
    return -1;
  }
  
  hprintf ("Disk drive detected: ");
  ide_fix_string (name, (unsigned char *)&buf[27], 40);
  hprintf ("%s", name);
  ide_fix_string (name, (unsigned char *)&buf[23], 8);
  hprintf ("%s", name);
  ide_fix_string (name, (unsigned char *)&buf[10], 20);
  hprintf ("%s", name);
  hprintf ("\n");
  
  return 0;
}

static int ide_idle (int dev)
{
  int ret;
  
  ret = ide_device_selection (0);
  if (ret) {
    return ret;
  }
  
  ret = wait_ready ();
  if (ret) {
    return ret;
  }
  
  ide_outb (0x00, IDE_SECTOR_COUNT);
  ide_outb (IDE_COMMAND_IDLE, IDE_COMMAND);
  delay (256);
  
  return wait_nbusy_ndrq ();
}

int ide_detect_devices ()
{
  int ret;

  ret = ide_reset ();
  if (ret) {
    return ret;
  }
  
  ret = ide_device_selection (0);
  if (ret) {
    return ret;
  }
  
  ret = ide_register_check ();
  if (ret) {
    return ret;
  }
  
  return ide_identify_device (0);
}

int ide_startup_devices (void)
{
  return ide_idle (0);
}

int ide_read_sectors (int dev, unsigned long lba,
		      unsigned char *buf, int count)
{
  int ret;
  
  ret = ide_device_selection (dev);
  if (ret) {
    return ret;
  }
  
  ret = wait_ready ();
  if (ret) {
    return ret;
  }
  
  ide_outb (count, IDE_SECTOR_COUNT);
  ide_outb (lba & 0xff, IDE_SECTOR_NUMBER);
  ide_outb ((lba & 0xff00) >> 8, IDE_SECTOR_CYLINDER_LOW);
  ide_outb ((lba & 0xff0000) >> 16, IDE_SECTOR_CYLINDER_HIGH);
  ide_outb (IDE_LBA | (dev << 4) | ((lba & 0xf000000) >> 24), IDE_DEVICE_HEAD);
  
  ide_outb (IDE_COMMAND_READ_SECTORS, IDE_COMMAND);
  delay (256);
  
  return ide_get_data (buf, count);
}
